'''
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''
from resource_management.libraries.script.script import Script
from resource_management.core.exceptions import Fail
from resource_management.libraries.resources.xml_config import XmlConfig
from resource_management.core.source import InlineTemplate
from resource_management.core.exceptions import ComponentIsNotRunning
from resource_management.core.logger import Logger
from resource_management.libraries.functions import format
from resource_management.libraries.functions import check_process_status
from resource_management.core.resources.system import File, Execute, Directory
from resource_management.core.resources.accounts import User
from resource_management.libraries.functions.default import default
import os
import os.path
import string
import random
import ConfigParser
import subprocess
import getpass
import re
import utils
import time
import sys


class Activity(Script):

    ##################### Directly exposed functions through Ambari #####################
    def __init__(self, component=None):
        self.component = component

    def reinstall(self, env):
        self.uninstall(env)
        self.install(env)

    def install(self, env):
        import params
        env.set_params(params)

        self.install_packages()

        if self.component is 'analyzer':
            self.install_activity_analyzer()
        elif self.component is 'explorer':
            self.install_activity_explorer()

        self.deploy_component_specific_config(env)


    # Configuration setting based on configs specified in ambari
    def configure(self, env):
        self.deploy_component_specific_config(env)


    # Start all SmartSense services
    def start(self, env):
        # Before start deploy the latest configs
        self.deploy_component_specific_config(env)
        if self.component is 'analyzer':
            self.start_activity_analyzer()
        elif self.component is 'explorer':
            self.start_activity_explorer()

    # Stop all SmartSense services
    def stop(self, env):
        if self.component is 'explorer':
            self.stop_activity_explorer()
        elif self.component is 'analyzer':
            self.stop_activity_analyzer()


    def status(self, env):
        import params
        if self.component is 'explorer':
            cmd = '/usr/sbin/hst activity-explorer status '
            exit_code, output, error = self.execute_command(cmd, failOnError=False)
            if (exit_code == 0 or exit_code == None):
                pass
            else:
                raise ComponentIsNotRunning()
        elif self.component is 'analyzer':
            # use built-in method to check status using pidfile
            print ("Pid file: " + str(params.activity_analyzer_pid_file))
            check_process_status(params.activity_analyzer_pid_file)

    def delete(self, env):
        self.uninstall(env)
        
    def uninstall(self, env):
        if self.component is 'analyzer':
            self.uninstall_activity_analyzer()
        elif self.component is 'explorer':
            self.uninstall_activity_explorer()


    def start_activity_analyzer(self):
        cmd = '/usr/sbin/hst activity-analyzer start '
        exit_code, output, error = self.execute_command(cmd)

    def stop_activity_analyzer(self):
        cmd = '/usr/sbin/hst activity-analyzer stop '
        exit_code, output, error = self.execute_command(cmd)

    def start_activity_explorer(self):
        import params
        cmd = 'export JAVA_HOME=' + params.config['hostLevelParams']['java_home'] + '; /usr/sbin/hst activity-explorer start '
        exit_code, output, error = self.execute_command(cmd)

    def stop_activity_explorer(self):
        cmd = '/usr/sbin/hst activity-explorer stop '
        exit_code, output, error = self.execute_command(cmd)

    def install_packages(self):
        import params
        import time
        distname, version = utils.get_os()
        major_version = version.split(".")[0]

        if distname.startswith('ubuntu') or distname.startswith('debian'):
            cmd = "{sudo} dpkg-query -l  | grep  'ii\s*smartsense-*' || {sudo} apt-get -o Dpkg::Options::=--force-confdef --allow-unauthenticated --assume-yes install smartsense-hst || {sudo} dpkg -i " + os.path.join(params.service_package_folder, "files" , "deb", "*.deb")
        elif distname.startswith('sles') or distname.startswith('suse'):
            cmd = "{sudo} rpm -qa | grep smartsense- || {sudo} zypper install --auto-agree-with-licenses --no-confirm smartsense-hst || {sudo} rpm -i " + os.path.join(params.service_package_folder, "files" , "rpm", "*.rpm")
        else:
            cmd = "{sudo} rpm -qa | grep smartsense- || {sudo} yum -y install smartsense-hst || {sudo} rpm -i " + os.path.join(params.service_package_folder, "files" , "rpm", "*.rpm")
        print("installing using command: " + cmd)

        attempts = 0
        while attempts < 10:
            attempts += 1
            try:
                exit_code, output, error = self.execute_command(cmd)
                if (exit_code == 0):
                    break
            except Exception, e:
                print "Failed to install during attempt " + str (attempts)
                print e
            if (attempts < 15):
                print "Waiting 10 seconds for next retry"
                time.sleep(5)

    def install_activity_explorer(self):
        import params
        user, group = self.get_owner_user_and_group()
        cmd = "{sudo} /usr/sbin/hst activity-explorer setup " + user + ":" + group
        print("Deploying activity explorer")
        exit_code, output, error = self.execute_command(cmd)

    def install_activity_analyzer(self):
        import params
        
        User(params.config['configurations']['activity-conf']['global.activity.analyzer.user'],
          gid=params.config['configurations']['cluster-env']['user_group'],
          groups=params.config['configurations']['hadoop-env']['hdfs_user']
         )
        
        user, group = self.get_owner_user_and_group()
        init_d_dir = "/etc/rc.d/init.d"
        distname, version = utils.get_os()
        if distname.startswith('sles') or distname.startswith('suse'):
            init_d_dir = "/etc/init.d"
        cmd = "{sudo} /usr/sbin/hst activity-analyzer setup " + user + ":" + group + " '" + init_d_dir + "'"
        print("Deploying activity analyzer")
        exit_code, output, error = self.execute_command(cmd)
        
        
        
        
    def uninstall_activity_explorer(self):
        import params
        cmd = "{sudo} " + params.hst_install_home + "/bin/uninstall-activity-explorer.sh "
        print("Removing activity explorer")
        exit_code, output, error = self.execute_command(cmd, failOnError=False)

    def uninstall_activity_analyzer(self):
        import params
        cmd = "{sudo} " + params.hst_install_home + "/bin/uninstall-activity-analyzer.sh "
        print("Removing activity analyzer")
        exit_code, output, error = self.execute_command(cmd, failOnError=False)

    def deploy_component_specific_config(self, env):
        import params
        env.set_params(params)

        # Before changing configs make sure it has right permissions to read
        self.set_permissions()

        # Additionally if it is a analyzer also deploy analyzer configs
        if self.component is 'explorer':
            self.configure_activity_explorer(env)
        elif self.component is 'analyzer':
            conf_file_name = params.activity_analyzer_conf_file_name
            conf_params = params.activity_analyzer_config
            self.deploy_ini_config(conf_file_name, conf_params)
            
            # Write env
            user, group = self.get_owner_user_and_group()
            if ('activity-env' in params.config['configurations'] and 'activity-env-content' in params.config['configurations']['activity-env']):
                # write out activity-env.sh
                activity_env_content = InlineTemplate(params.config['configurations']['activity-env']['activity-env-content'])
                File(params.activity_conf_dir + "/activity-env.sh", content=activity_env_content,
                     owner=user, group=group)
      
            
        # Write log4j
        log4j_props = None
        if ('activity-log4j' in params.config['configurations']) and ('activity-log4j-content' in params.config['configurations']['activity-log4j']):
            log4j_props = InlineTemplate(params.config['configurations']['activity-log4j']['activity-log4j-content'])
        if (log4j_props != None):
            File(format(params.activity_log4j_conf_file),
            mode=0644,
            content=log4j_props
            )
        # If this is not AMS collector, make sure this host has necessary AMS configs
        if (params.is_ams_hbase_distributed == True and  self.am_i_metric_collector() == False ):
            print("This host is not running metrics collector so explicitly deploying ams-hbase-conf that is needed by " + self.component)
            
            self.execute_command("{sudo} mkdir -p " + params.ams_hbase_conf_dir)
            self.execute_command("{sudo} chmod -R 755 " + params.ams_hbase_conf_dir + "/..")
            self.execute_command("{sudo} chown -R " + params.ams_user + ":" + params.ams_user_group + " " + params.ams_hbase_conf_dir + "/..")
            if ('configurations' in params.config and 'hdfs-site' in  params.config['configurations']):
                XmlConfig("hdfs-site.xml",
                    conf_dir=params.ams_hbase_conf_dir,
                    configurations=params.config['configurations']['hdfs-site'],
                    configuration_attributes=params.config['configuration_attributes']['hdfs-site'],
                    owner=params.ams_user,
                    group=params.ams_user_group,
                    mode=0644
                    )
    
            if ('configurations' in params.config and 'core-site' in  params.config['configurations']):
                XmlConfig("core-site.xml",
                    conf_dir=params.ams_hbase_conf_dir,
                    configurations=params.config['configurations']['core-site'],
                    configuration_attributes=params.config['configuration_attributes']['core-site'],
                    owner=params.ams_user,
                    group=params.ams_user_group,
                    mode=0644
                )
    
            if ('configurations' in params.config and 'ams-hbase-site' in  params.config['configurations']):
                
                merged_ams_hbase_site = {}
                if ('ams-hbase-site' in params.config['configurations']):
                    merged_ams_hbase_site.update(params.config['configurations']['ams-hbase-site'])
                if (params.is_ams_hbase_secured == True and 'ams-hbase-security-site' in params.config['configurations']) :
                    merged_ams_hbase_site.update(params.config['configurations']['ams-hbase-security-site'])

                # Add phoenix client side overrides
                merged_ams_hbase_site['phoenix.query.maxGlobalMemoryPercentage'] = str(params.ams_phoenix_max_global_mem_percent)
                merged_ams_hbase_site['phoenix.spool.directory'] = params.ams_phoenix_client_spool_dir
                merged_ams_hbase_site['hbase.zookeeper.quorum'] = params.phoenix_zk_quorum
                merged_ams_hbase_site['hbase.zookeeper.property.clientPort'] = params.phoenix_zk_port
                merged_ams_hbase_site['zookeeper.znode.parent'] = params.phoenix_zk_node
                merged_ams_hbase_site['hbase.tmp.dir'] = '/var/lib/smartsense/activity-analyzer/tmp/phoenix'
    
                
                XmlConfig("hbase-site.xml",
                    conf_dir=params.ams_hbase_conf_dir,
                    configurations=merged_ams_hbase_site,
                    configuration_attributes=params.config['configuration_attributes']['ams-hbase-site'],
                    owner=params.ams_user,
                    group=params.ams_user_group,
                    mode=0644
                )
        # After changing configs make sure it has right permissions
        self.set_permissions()


    def configure_activity_explorer(self, env):
        import params
        env.set_params(params)
        
        if ('configurations' in params.config):
            user, group = self.get_owner_user_and_group()
            if ('activity-zeppelin-site' in params.config['configurations']):
                # write out zeppelin-site.xml
                XmlConfig("zeppelin-site.xml",
                          conf_dir=params.activity_conf_dir,
                          configurations=params.config['configurations']['activity-zeppelin-site'],
                          owner=user,
                          group=group
                          )
            if ('activity-zeppelin-env' in params.config['configurations'] and 'activity-zeppelin-env-content' in params.config['configurations']['activity-zeppelin-env']):
                # write out zeppelin-env.sh
                zeppelin_env_content = InlineTemplate(params.config['configurations']['activity-zeppelin-env']['activity-zeppelin-env-content'])
                File(params.activity_conf_dir + "/zeppelin-env.sh", content=zeppelin_env_content,
                     owner=user, group=group)
    
            if ('activity-zeppelin-interpreter' in params.config['configurations'] and 'activity-zeppelin-interpreter-content' in params.config['configurations']['activity-zeppelin-interpreter']):
                # write out interpreter.json
                zeppelin_interpreter_content = InlineTemplate(params.config['configurations']['activity-zeppelin-interpreter']['activity-zeppelin-interpreter-content'])
                File(params.activity_conf_dir + "/interpreter.json", content=zeppelin_interpreter_content,
                     owner=user, group=group)

            if ('activity-zeppelin-shiro' in params.config['configurations']):
                # write out shiro.ini
                self.deploy_ini_config(params.activity_conf_dir + "/shiro.ini", params.config['configurations']['activity-zeppelin-shiro'])

    def deploy_ini_config(self, conf_file_name, conf_params):
        import params
        import os

        config_file_path = os.path.join(params.activity_conf_dir, conf_file_name)

        # Derive java home from ambari and set it
        java_home = params.config['hostLevelParams']['java_home']
        # print("java_home="+str(java_home))

        config = ConfigParser.ConfigParser()
        config.optionxform = str       
        config.read(config_file_path)
        
        if config_file_path.endswith("activity.ini") :
            if (not config.has_section('java')):
                config.add_section('java')
            config.set('java', 'home', java_home)

            if ((not config.has_option('analyzer', 'run.as.user')) or config.get('analyzer', 'run.as.user') == None or config.get('analyzer', 'run.as.user').strip == ""):
                if ("root" != getpass.getuser()):
                    if (not config.has_section('analyzer')):
                        config.add_section('analyzer')
                    config.set('analyzer', 'run.as.user', getpass.getuser())
            if params.activity_analyzer_jdbc_url != None:
                if (not config.has_section('phoenix')):
                    config.add_section('phienix')
                config.set('phoenix', 'activity.analyzer.jdbc.url', params.activity_analyzer_jdbc_url)
            
            if params.ams_jdbc_url != None:
                if (not config.has_section('ams')):
                    config.add_section('ams')
                config.set('ams', 'jdbc.url', params.ams_jdbc_url)
            
            if (not config.has_section('customer')):
                config.add_section('customer')
            config.set('customer', 'smartsense.id', params.hst_server_config['customer.smartsense.id'])

            if (not config.has_section('cluster')):
                config.add_section('cluster')
            config.set('cluster', 'name', params.config['clusterName'])

        if conf_params != None:
            for k, v in conf_params.iteritems():
                # print("Setting server config " + str(k) + '=' + str(v))
                key = k.split(".", 1)
                if (len(key) != 2):
                    raise Fail("Invalid property " + k + ". Configuration property should be named in the form <section name>.<property name>")
                if (str(v).lower() == "null") and config.has_option(key[0], key[1]):
                    config.remove_option(key[0], key[1])
                else:
                    if (not config.has_section(key[0])):
                        config.add_section (key[0])
                    config.set(key[0], key[1], v)

        print("Writing configs to : " + str(config_file_path))

        # First write to temp file and then rename to avoid partial written files.
        with open(config_file_path + ".tmp", 'w') as config_file:
            config.write(config_file)
        os.rename(config_file_path + ".tmp", config_file_path)


    def set_permissions(self):
        import params
        dir_list = []
        user, group = self.get_owner_user_and_group()

        if self.component is 'explorer':
            dir_list = params.activity_explorer_owned_dirs
            exit_code, output, error = self.execute_command("{sudo} chown '" + user + "':'" + group + "' " + params.activity_explorer_keytab, printOutput=False, failOnError=False)
        elif self.component is 'analyzer':
            dir_list = params.activity_analyzer_owned_dirs
            analyzer_config = ConfigParser.ConfigParser()
            analyzer_config.optionxform = str
            analyzer_config.read(os.path.join(params.activity_conf_dir, params.activity_analyzer_conf_file_name))
            exit_code, output, error = self.execute_command("{sudo} chown '" + user + "':'" + group + "' " + params.activity_analyzer_keytab, printOutput=False, failOnError=False)
            if 'activity-conf' in params.config['configurations'] and 'global.activity.analyzer.user.keytab' in params.config['configurations']['activity-conf']:
                exit_code, output, error = self.execute_command("{sudo} chown '" + user + "':'" + group + "' " + params.config['configurations']['activity-conf']['global.activity.analyzer.user.keytab'], failOnError=False)
            if 'activity-conf' in params.config['configurations'] and 'mr_job.activity.analyzer.user.keytab' in params.config['configurations']['activity-conf']:
                exit_code, output, error = self.execute_command("{sudo} chown '" + user + "':'" + group + "' " + params.config['configurations']['activity-conf']['mr_job.activity.analyzer.user.keytab'], failOnError=False)
            if 'activity-conf' in params.config['configurations'] and 'tez_job.activity.analyzer.user.keytab' in params.config['configurations']['activity-conf']:
                exit_code, output, error = self.execute_command("{sudo} chown '" + user + "':'" + group + "' " + params.config['configurations']['activity-conf']['tez_job.activity.analyzer.user.keytab'], failOnError=False)
            if 'activity-conf' in params.config['configurations'] and 'yarn_app.activity.analyzer.user.keytab' in params.config['configurations']['activity-conf']:
                exit_code, output, error = self.execute_command("{sudo} chown '" + user + "':'" + group + "' " + params.config['configurations']['activity-conf']['yarn_app.activity.analyzer.user.keytab'], failOnError=False)
            
        for dir in dir_list :
            cmd = "{sudo} mkdir -p " + dir
            exit_code, output, error = self.execute_command(cmd, failOnError=False)

            cmd = "{sudo} chown -R '" + user + "':'" + group + "' " + dir
            exit_code, output, error = self.execute_command(cmd, failOnError=False)


    def get_owner_user_and_group(self):
        import params
        run_as_user = ""
        if self.component is 'analyzer':
            analyzer_config = ConfigParser.ConfigParser()
            analyzer_config.optionxform = str
            analyzer_config.read(os.path.join(params.activity_conf_dir, params.activity_analyzer_conf_file_name))
            if (analyzer_config.has_section('analyzer') and analyzer_config.has_option('analyzer', 'run.as.user')):
               run_as_user = analyzer_config.get('analyzer', 'run.as.user')

        if (run_as_user == None or run_as_user.strip() == ""):
            run_as_user = getpass.getuser()
        group = run_as_user
        id_cmd = 'id -gn ' + run_as_user
        p = subprocess.Popen([id_cmd], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True)
        output, error = p.communicate()
        if (p.returncode == 0 or p.returncode == None) :
            group = output.strip()
        else:
            group = output.user()
        return (run_as_user, group)


    def get_sudo_command(self):
        if os.geteuid() == 0:
            return ""
        else:
            return "sudo"

    def execute_command(self, command, background=False, failOnError=True, debug=False, printOutput=True):
        normalized_command = command.replace("{sudo}", self.get_sudo_command())
        if (debug == True):
            print ("Command to be executed:" + command)
            print ("Normalized command:" + normalized_command)
    
        if (background == True):
            p = subprocess.Popen(normalized_command, shell=True)
            # Wait for process to spawn
            time.sleep(3)
            return (0, None, None)
    
        p = subprocess.Popen(normalized_command, stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE, shell=True)
        stdout, stderr = p.communicate()
        code = p.wait()
        if (debug == True or printOutput == True):
            print("Command: " + str(normalized_command))
            print("Exit code: " + str(code))
            print("Std Out: " + ("None" if stdout == None or len(str(stdout).strip()) == 0 else str(stdout)))
            print("Std Err: " + ("None" if stderr == None or len(str(stderr).strip()) == 0 else str(stderr)))
        if (failOnError == True and code != 0):
            sys.stderr.write("Failed to execute command: " + str(normalized_command) + "; Exit code: " + str(code) + "; stdout: " + str(stdout) + "; stderr: " + str(stderr))
            sys.stderr.flush()
            sys.exit(1)
        return (code, stdout, stderr)
 
    def am_i_metric_collector(self):
        import params
        current_host = params.config['hostname']
        if current_host in params.config['clusterHostInfo']['metrics_collector_hosts']:
            return True
        else:
            return False


if __name__ == "__main__":
    Activity().execute()
